/*
**                      Softwarepraktikum iMaze
**                              1993/94
**                Joerg Czeranski    Hans-Ulrich Kiel
**
** Datei: X_tasten.c
**
** Kommentar:
**  Wertet Events von der Tastatur aus und setzt definierte Signale
*/


#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>

#ifdef JOYSTICK
#ifdef linux

#include <linux/joystick.h>

#ifndef DEFAULT_JOYSTICK_DEVICE
#define DEFAULT_JOYSTICK_DEVICE "/dev/js0"
#endif

#undef USE_SELECT
#undef CHECK_MAX

#else
#ifdef __FreeBSD__

#include <machine/joystick.h>

#ifndef DEFAULT_JOYSTICK_DEVICE
#define DEFAULT_JOYSTICK_DEVICE "/dev/joy0"
#endif

#define USE_SELECT
#define CHECK_MAX

#else

#undef JOYSTICK

#endif
#endif
#endif

#include <X11/Xlib.h>
#include <X11/keysym.h>

#include "global.h"
#include "system.h"
#include "signale.h"
#include "ereignisse.h"
#include "einaus.h"
#include "grafik.h"

static char sccsid[] = "@(#)X_tasten.c	2.12 02/22/96";


#define GEDRUECKT 1       /* Bit 0 wird gesetzt, wenn Taste gedrueckt;
                             und geloescht, wenn diese losgelassen */
#define NICHT_ABGEFRAGT 2 /* Bit 1 bleibt gesetzt, bis ein Tastendruck
                             registriert wurde */

/* die Cursor-Tasten haben eine Repeat-Funktion, alle anderen nicht */
#define HAT_REPEAT(signalnr) \
	((signalnr) == VORSIGNAL || (signalnr) == ZURUECKSIGNAL || \
	(signalnr) == LINKSSIGNAL || (signalnr) == RECHTSSIGNAL)

static char tasten_aktiv[SIGNALANZ]; /* aktueller Tastenzustand
                                        (GEDRUECKT / NICHT_ABGEFRAGT) */

#ifdef JOYSTICK
static char joystick_aktiv[SIGNALANZ]; /* aktueller Joystickzustand
                                          (GEDRUECKT / NICHT_ABGEFRAGT) */

static int joystick;          /* Deskriptor fuer Joystick-Device */
static int joystick_abfragen; /* Flag: Joystick benutzen? */
static int joystick_gelesen;  /* Flag: joystick_xmax, ... initialisiert? */
static int joystick_xmin;     /* minimaler x-Wert der Joystick-Position */
static int joystick_xmax;     /* maximaler x-Wert der Joystick-Position */
static int joystick_ymin;     /* minimaler y-Wert der Joystick-Position */
static int joystick_ymax;     /* maximaler y-Wert der Joystick-Position */

#ifdef CHECK_MAX
static int joystick_maxval; /* Wert, falls Abfrage ergebnislos (Timeout) */
#endif

#endif


#ifdef JOYSTICK
/*
** fehler_text
**  ermittelt den zur Systemvariable errno passenden Fehlertext
**
** Rueckgabewert:
**  Zeiger in das Feld sys_errlist oder in eine lokale Variable, die
**  den Fehler als Nummer enthaelt (falls kein Standardtext vorhanden ist)
**
** Seiteneffekte:
**  der zuletzt zurueckgegebene Zeiger wird evtl. ungueltig
*/
static char *fehler_text()
{
	extern int errno;           /* Systemvariable: letzter aufgetretener
	                               Fehlercode oder 0 */
#ifndef DONT_DECLARE_ERRLIST
	extern char *sys_errlist[]; /* Liste der Standard-Fehlertexte */
	extern int sys_nerr;        /* Anzahl der Eintraege in sys_errlist */
#endif

	/* falls errno kein gueltiger Index fuer sys_errlist ist, nur die
	   Nummer als String zurueckgeben */
	if (errno < 0 || errno >= sys_nerr)
	{
		static char text[20]; /* Puffer fuer Fehlernummer */

		sprintf(text, "Error #%d", errno);

		/* Zeiger auf Puffer zurueckgeben; Achtung: dieser Zeiger
		   ist immer derselbe */
		return text;
	}

	/* Standard-Fehlertext zurueckgeben */
	return (char *)sys_errlist[errno];
}
#endif


/*
** aktiv_setzen
**  setzt oder loescht das "gedrueckt"-Flag fuer ein Signal
**
** Parameter:
**  aktiv: Feld mit Flags fuer die Signale
**  signalnr: Nummer des Signals
**  gedrueckt: 1, falls Taste gedrueckt wurde; 0, falls losgelassen
**
** Seiteneffekte:
**  aktiv[signalnr] wird veraendert
*/
static void aktiv_setzen(aktiv, signalnr, gedrueckt)
char aktiv[SIGNALANZ];
int signalnr, gedrueckt;
{
	/* signalnr ungueltig? */
	if (signalnr < 0 || signalnr >= SIGNALANZ)
		return;

	/* Taste wurde losgelassen */
	if (!gedrueckt)
		aktiv[signalnr] &= ~GEDRUECKT;
	/* Taste ist niedergedrueckt (neu oder durch repeat);
	   falls Taste nicht zuvor schon gedrueckt war, als
	   gedrueckt und nicht abgefragt markieren */
	else if (!(aktiv[signalnr] & GEDRUECKT))
		aktiv[signalnr] = GEDRUECKT | NICHT_ABGEFRAGT;
}


/*
** X_event_routine
**  verarbeitet die Tastatur-Events und ordnet sie den internen Signalen zu
**
** Parameter:
**  event: Beschreibung des Events
**
** Seiteneffekte:
**  aktiv wird veraendert
*/
void X_event_routine(display, event)
Display *display;
XEvent *event;
{
	int signalnr;  /* Index auf Tabelle mit Signalen */
	KeySym action; /* Typ bzw. ASCII-Code des Events */

	switch (event->type)
	{
		case FocusOut: /* Fenster verliert Fokus (erhaelt keine
		                  Tastatur-Eingaben mehr) */
			/* dann werden alle Tasten wie losgelassen behandelt */
			for (signalnr = 0; signalnr < SIGNALANZ; signalnr++)
				tasten_aktiv[signalnr] &= ~GEDRUECKT;
			return;
		case KeyPress:
		case KeyRelease:
			break;
		default:
			return;
	}

	switch (action = XKeycodeToKeysym(display, event->xkey.keycode, 0))
	{
		case XK_c: /* Control-C */
			if (event->type == KeyPress &&
				!(event->xkey.state & ControlMask))
				return;
			if (event->type == KeyPress)
				disconnect_abfrage();
			return;
		case XK_q: /* Control-Q */
			if (event->type == KeyPress &&
				!(event->xkey.state & ControlMask))
				return;
			signalnr = WEITERSIGNAL;
			break;
		case XK_s: /* Control-S */
			if (event->type == KeyPress &&
				!(event->xkey.state & ControlMask))
				return;
			signalnr = STOPSIGNAL;
			break;
		case XK_space: /* Leertaste */
		case XK_Shift_L:
		case XK_Shift_R:
		case XK_Alt_L:
		case XK_Alt_R:
			signalnr = SCHUSSSIGNAL;
			break;
		case XK_Up: /* Cursor hoch */
			signalnr = VORSIGNAL;
			break;
		case XK_Down: /* Cursor runter */
		case XK_KP_5: /* 5 auf dem Ziffernblock */
			signalnr = ZURUECKSIGNAL;
			break;
		case XK_Left: /* Cursor links */
			signalnr = LINKSSIGNAL;
			break;
		case XK_Right: /* Cursor rechts */
			signalnr = RECHTSSIGNAL;
			break;
		case XK_0: /* Tasten 0 bis 9 */
			signalnr = AKTIONSSIGNALE + 9;
			break;
		case XK_1:
		case XK_2:
		case XK_3:
		case XK_4:
		case XK_5:
		case XK_6:
		case XK_7:
		case XK_8:
		case XK_9:
			signalnr = AKTIONSSIGNALE + action - XK_1;
			break;
		default:
			return;
	}

	/* Tastenflags setzen */
	aktiv_setzen(tasten_aktiv, signalnr, event->type == KeyPress);
}


void eingabe_simulieren(signalnr)
int signalnr;
{
	/* signalnr ungueltig? */
	if (signalnr < 0 || signalnr >= SIGNALANZ)
		return;

	tasten_aktiv[signalnr] |= NICHT_ABGEFRAGT;
}


/* bis hier lokaler Teil                       */
/***********************************************/
/* ab hier globaler Teil                       */


/*
** eingabe_abfragen
**  setzt in einem Feld aller Signale diejenigen, die seit der letzten
**  Abfrage aufgetreten sind. Die Tasten werden als abgefragt markiert
**
** Paramter:
**  signale: Feld mit allen Signalen
**           (0 = nicht aufgetreten; 1 = aufgetreten)
**
** Seiteneffekte:
**  aktiv und signale wird veraendert
*/
void eingabe_abfragen(signale)
char signale[SIGNALANZ];
{
	int signalnr; /* Index auf Tabelle mit Signalen */

#ifdef JOYSTICK
	/* Joystick benutzen? */
	if (joystick_abfragen)
	{
		int laenge;                         /* Laenge der gelesenen Daten */
		int j_x, j_y, j_button;             /* gelesene Daten */
#ifdef linux
		struct JS_DATA_TYPE joystick_daten; /* Position des Joysticks */
#else
		struct joystick joystick_daten;
#endif

#ifdef USE_SELECT
		fd_set deskriptoren;
		struct timeval timeout;

		FD_ZERO(&deskriptoren);
		FD_SET(joystick, &deskriptoren);

		timeout.tv_sec = timeout.tv_usec = 0;

		if (select(max_deskriptor_anzahl(), &deskriptoren, NULL,
			NULL, &timeout) > 0)
			laenge = read(joystick, &joystick_daten, sizeof joystick_daten);
		else
			laenge = 0;
#else
		/* Joystick abfragen */
		laenge = read(joystick, &joystick_daten, sizeof joystick_daten);
#endif

		/* Fehler? */
		if (laenge < 0 && errno != EWOULDBLOCK && errno != EAGAIN)
		{
			static char *meldung[] = { "iMaze - Joystick Error", "",
				"Can't read from joystick device:", NULL, NULL };

			meldung[3] = fehler_text();
			uebler_fehler(meldung, NULL);
		}

		/* stimmt die Laenge der gelesenen Daten? */
		if (laenge == sizeof joystick_daten)
		{
			j_x = joystick_daten.x;
			j_y = joystick_daten.y;

#ifdef CHECK_MAX
			if (j_x < 0)
				j_x = joystick_maxval;
			if (j_y < 0)
				j_y = joystick_maxval;
#endif

#ifdef linux
			j_button = joystick_daten.buttons != 0;
#else
			j_button = joystick_daten.b1 || joystick_daten.b2;
#endif

			/* xmin, ... schon initialisiert? */
			if (joystick_gelesen)
			{
				joystick_xmin = min(j_x, joystick_xmin);
				joystick_xmax = max(j_x, joystick_xmax);
				joystick_ymin = min(j_y, joystick_ymin);
				joystick_ymax = max(j_y, joystick_ymax);
			}
			else
			{
				joystick_xmin = joystick_xmax = j_x;
				joystick_ymin = joystick_ymax = j_y;

				joystick_gelesen = 1;
			}

			/* Joystick-Buttons abfragen */
			aktiv_setzen(joystick_aktiv, SCHUSSSIGNAL, j_button);

			/* x-Richtung abfragen */
			aktiv_setzen(joystick_aktiv, LINKSSIGNAL,
				3 * j_x < 2 * joystick_xmin + joystick_xmax);
			aktiv_setzen(joystick_aktiv, RECHTSSIGNAL,
				3 * j_x > joystick_xmin + 2 * joystick_xmax);

			/* y-Richtung abfragen */
			aktiv_setzen(joystick_aktiv, VORSIGNAL,
				3 * j_y < 2 * joystick_ymin + joystick_ymax);
			aktiv_setzen(joystick_aktiv, ZURUECKSIGNAL,
				3 * j_y > joystick_ymin + 2 * joystick_ymax);
		}
	}
#endif

	for (signalnr = 0; signalnr < SIGNALANZ; signalnr++)
	{
		/* falls REPEAT-Funktion vorgesehen (Cursor-Tasten),
		   Signal setzen, wenn GEDRUECKT oder NICHT_ABGEFRAGT,
		   sonst nur wenn NICHT_ABGEFRAGT */
		if (HAT_REPEAT(signalnr))
			signale[signalnr] = tasten_aktiv[signalnr] != 0;
		else
			signale[signalnr] =
				(tasten_aktiv[signalnr] & NICHT_ABGEFRAGT) != 0;

		/* alle Tasten als ABGEFRAGT markieren */
		tasten_aktiv[signalnr] &= ~NICHT_ABGEFRAGT;

#ifdef JOYSTICK
		/* Joystick benutzen? */
		if (joystick_abfragen)
		{
			/* falls REPEAT-Funktion vorgesehen (Joystick-Bewegung),
			   Signal setzen, wenn GEDRUECKT oder NICHT_ABGEFRAGT,
			   sonst nur wenn NICHT_ABGEFRAGT */
			if (HAT_REPEAT(signalnr))
				signale[signalnr] |= joystick_aktiv[signalnr] != 0;
			else
				signale[signalnr] |=
					(joystick_aktiv[signalnr] & NICHT_ABGEFRAGT) != 0;

			/* alle Joystick-Daten als ABGEFRAGT markieren */
			joystick_aktiv[signalnr] &= ~NICHT_ABGEFRAGT;
		}
#endif
	}
}


/*
** X_eingabe_init
**  initialisiert die Event-Abfrage
**
** Parameter:
**
** Rueckgabewert:
**  0 fuer Erfolg, 1 fuer Fehler
*/
int X_eingabe_init(argc, argv)
int *argc;
char **argv;
{
	int i; /* Index */

#ifdef JOYSTICK
	char *joystick_device; /* Name des Joystick-Devices */
	int ueberspringen;     /* Anzahl der ausgewerteten
	                          Kommandozeilenparameter */

	/* initialisieren */
	joystick_abfragen = 0;
	joystick_device = DEFAULT_JOYSTICK_DEVICE;

	/* Kommandozeilenparameter auswerten */
	for (ueberspringen = 1; ueberspringen < *argc &&
		argv[ueberspringen][0] == '-'; ueberspringen++)
	{
		for (i = 1; argv[ueberspringen][i] != 0 &&
			strchr("j", argv[ueberspringen][i]) != NULL; i++);

		if (strchr("J", argv[ueberspringen][i]) != NULL)
		{
			if (argv[ueberspringen][i + 1] == 0 &&
				ueberspringen == *argc - 1)
				break;
		}
		else if (argv[ueberspringen][i] != 0 || i < 2)
			break;

		/* argv[ueberspingen] faengt mit "-" an und enthaelt
		   nur bekannte Optionen; Optionen auswerten */
		for (i = 1; i > 0 && argv[ueberspringen][i] != 0;)
			switch (argv[ueberspringen][i++])
			{
				case 'J': /* Joystick benutzen mit alternativem Device */
					if (argv[ueberspringen][i] == 0)
						ueberspringen++, i = 0;
					joystick_device = argv[ueberspringen] + i;
					i = 0;

				case 'j': /* Joystick benutzen */
					joystick_abfragen = 1;
			}
	}

	/* "--" ignorieren */
	if (ueberspringen < *argc && !strcmp(argv[ueberspringen], "--"))
		ueberspringen++;

	/* mehr als argv[0] ausgewertet? */
	if (ueberspringen > 1)
	{
		/* Aufrufparameter verschieben, ausgewertete ueberschreiben */
		ueberspringen--;
		*argc -= ueberspringen;
		for (i = 1; i < *argc; i++)
			argv[i] = argv[i + ueberspringen];
	}

	/* Joystick nicht benutzen? */
	if (!joystick_abfragen)
		return 0;

	/* Joystick-Device oeffnen Erfolg oder Abbruch */
	for (;;)
	{
		static char *meldung[] = { "iMaze - Joystick Error",
			"", NULL, NULL, NULL };

		/* Joystick-Device oeffnen */
		if ((joystick = open(joystick_device, O_RDONLY | O_NDELAY, 0)) >= 0)
			break;

		/* Fehlertext erzeugen */
		speicher_belegen((void **)&meldung[2],
			strlen(joystick_device) + 20);
		sprintf(meldung[2], "Can't open %s:", joystick_device);
		meldung[3] = fehler_text();

		/* sonstiger Fehler? */
		if (errno != EBUSY)
			if (rueckfrage(meldung, "No Joystick", NULL))
			{
				joystick_abfragen = 0;
				speicher_freigeben((void **)&meldung[2]);

				return 0;
			}
			else
			{
				speicher_freigeben((void **)&meldung[2]);

				return 1;
			}
		/* Device gerade in Benutzung? */
		else if (!rueckfrage(meldung, "Try again", "No Joystick"))
		{
			joystick_abfragen = 0;
			speicher_freigeben((void **)&meldung[2]);

			return 0;
		}

		speicher_freigeben((void **)&meldung[2]);
	}

#ifndef USE_SELECT
	/* Joystick-Device in nicht blockierenden Modus schalten */
	if (fcntl(joystick, F_SETFL, O_NONBLOCK))
	{
		static char *meldung[] = { "iMaze - Joystick Error", "",
			"Can't switch joystick device to non-blocking mode:",
			NULL, NULL };

		meldung[3] = fehler_text();
		uebler_fehler(meldung, NULL);
	}
#endif

#ifdef CHECK_MAX
	joystick_maxval = 10000; /* 10ms Timeout (Pfusch!) */

	/* Joystick-Device Timeout setzen */
	if (ioctl(joystick, JOY_SETTIMEOUT, &joystick_maxval))
	{
		static char *meldung[] = { "iMaze - Joystick Error", "",
			"Can't set joystick device timeout:",
			NULL, NULL };

		meldung[3] = fehler_text();
		uebler_fehler(meldung, NULL);
	}
#endif

	/* xmin, xmax, ... noch nicht initialisiert */
	joystick_gelesen = 0;
#endif

	return 0;
}


/*
** X_eingabe_ende
**  Routine fuer das Schliessen von Eingabekanaelen etc.
*/
void X_eingabe_ende()
{
#ifdef JOYSTICK
	/* Joystick schliessen */
	if (joystick_abfragen && close(joystick))
	{
		static char *meldung[] = { "iMaze - Joystick Error", "",
			"Can't close joystick device:", NULL, NULL };

		meldung[3] = fehler_text();
		uebler_fehler(meldung, NULL);
	}

	/* Joystick nicht mehr abfragen */
	joystick_abfragen = 0;
#endif
}
